/* $Id: UnidenScanner.cpp,v 1.4 2007/02/04 11:35:27 olle Exp $
 *
 * Copyright (c) 2007, olle@toolcrypt.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *    * Neither the name of the <ORGANIZATION> nor the names of its contributors may
 *      be used to endorse or promote products derived from this software without 
 *      specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


#include "stdafx.h"
#include <string.h>

#include "UnidenScanner.h"


CUnidenScanner::CUnidenScanner(UINT port, UINT bps, UINT bits, UINT parity, UINT stopbits)
	: CReceiverCtrl(port, bps, bits, parity, stopbits)
{
	// set default read timeout
	m_port.SetTimeout(portReadTimeOut);
}

CUnidenScanner::~CUnidenScanner()
{
	// clean up after ourselves
	clearReadBuffer();
}

BOOL CUnidenScanner::getNextLine(char *buf, UINT bufLen, const char *expect)
{
	// clear buffer
	memset(buf, 0, bufLen);

	// get next line
	UINT nBytesRead = m_port.ReadLine(buf, bufLen);

	// if we're waiting for a specific line, ignore all others
	if (expect)
	{
		size_t l = strlen(expect);
		while (strncmp(buf, expect, l) != 0)
		{
			// don't ignore "ERR" message
			if (strcmp(buf, "ERR") == 0)
				break;

			// stop if we can't get any more lines
			nBytesRead = m_port.ReadLine(buf, bufLen);
			if (nBytesRead < 1)
				break;
		}
	}

	return (nBytesRead > 0);
}

BOOL CUnidenScanner::checkResponse(const char *data, const char *expect, const char *method)
{
	if (!data || !expect)
		return FALSE;

	// check for expected response
	if (strncmp(data, expect, (int)strlen(expect) != 0))
	{
#ifdef DEBUG
		// report error to user
		static char msg[4096];
		if (strcmp(data, "ERR") == 0 && strcmp(expect, "OK") != 0)
			_snprintf(msg, 4096, "Error in %s: Command \"%s\" not supported!", method, expect);
		else
			_snprintf(msg, 4096, "Error in %s: Expected \"%s\", got \"%s\".", method, expect, data);

		MessageBox(NULL, msg, "Receiver Control", MB_ICONERROR);
#endif
		return FALSE;
	}

	return TRUE;
}

void CUnidenScanner::clearReadBuffer()
{
	// read any buffered data
	char tmp[4096];
	m_port.ReadData(tmp, 4096);
}

void CUniden250::startScan()
{
	UINT mode = 0;
	char buf[128];

	// check what mode we're in
	sendData("MD\r", 3);
	getNextLine(buf, 128);
	if (checkResponse(buf, "MD", "CUniden250::startScan()"))
	{
		buf[4]=0;
		mode = atoi(buf+3);
	}
	else
		return;

	// if we're in trunking mode
	if (mode > 8)
	{
		// press the RESUME key
		mode = getResumeKeyCode();
	} else {

		// press the SCAN key
		mode = getScanKeyCode();
	}

	// do the keypress
	_snprintf(buf, 128, "KEY%0.2u\r", mode);
	sendData(buf, 6);
	getNextLine(buf, 128);
	checkResponse(buf, "OK", "CUniden250::startScan(KEY)");
}

void CUniden250::setFrequency(DOUBLE freq, BOOL narrow)
{
	// set frequency of receiver
	if (m_memory)
	{
		// make sure we're in manual mode
		setManualMode();

		// program memory location
		char buf[128];
		_snprintf(buf, 128, "PM%0.3u %08.0f\r", m_memory, freq*10000);
		sendData(buf, 15);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "C", "CUniden250::setFrequency(PM)"))
			return;
		
		// set memory lockout
		sendData("LON\r", 4);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(LON)"))
			return;

		// set modulation mode
		if (narrow)
			sendData("RM NFM\r", 7);
		else
			sendData("RM FM\r", 6);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(RM)"))
			return;

	} else { // if no memory location configured

		// set freq using VFO tune
		char buf[128];
		_snprintf(buf, 128, "RF%08.0f\r", freq*10000);
		sendData(buf, 11);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(RF)"))
			return;
	}
}

DOUBLE CUniden250::getFrequency()
{
	DOUBLE freq = 0;
	char buf[128];

	// read signal strength command
	sendData("SG\r", 3);
	getNextLine(buf, 128);

	// parse response
	if (buf[5] == 'F')
	{
		buf[14]=0;
		freq = ((DOUBLE)atoi(buf+6)/10000);
	} else {
		checkResponse(buf, "NG", "CUniden250::getFrequency(SG)");
	}

	return freq;
}

void CUniden250::setText(LPCSTR text)
{
	if (m_memory)
	{
		// set text for memory location in m_memory
		char buf[128];
		_snprintf(buf, 128, "TA C %0.3u %s\r", m_memory, text);
		sendData(buf, (UINT)strlen(buf));
		getNextLine(buf, 128);
		checkResponse(buf, "OK", "CUniden250::setText(TA)");
	}
}

void CUniden250::setText2(LPCSTR text)
{
	if (m_memory)
	{
		// set bank text
		char buf[128];
		_snprintf(buf, 128, "TA B %c %s\r", m_bank, text?text:(LPCSTR)m_banktag); // if text == NULL, set default bank tag
		sendData(buf, (UINT)strlen(buf));
		getNextLine(buf, 128);
		checkResponse(buf, "OK", "CUniden250::setText2(TA)");
	}
}


BOOL CUniden250::isSquelchOpen()
{
	// check squelch status
	char buf[50];
	sendData("SQ\r", 3);
	getNextLine(buf, 50);

	// check return
	switch(buf[0])
	{
	// indicate squelch setting
	case '+':
		return TRUE;
	case '-':
		return FALSE;

	default:
		checkResponse(buf, "SQ response", "CUniden250::isSquelchOpen()");
	}

	return FALSE;
}

void CUniden250::setBacklight(CReceiverCtrl::BacklightState state)
{
	// set the backlight
	char buf[64];
	_snprintf(buf, 64, "LT%s\r", (state!=Off)?((state==High)?"N":"D"):"F");
	sendData(buf, 4);
	getNextLine(buf, 64);
	checkResponse(buf, "OK", "CUniden250::setBacklight(LT)");
}

BOOL CUniden250::getBankTag()
{
	// get bank for m_memory
	m_bank = channelToBank(m_memory);

	// clear the read buffer
	clearReadBuffer();

	// get current bank text
	char buf[128];
	_snprintf(buf, 128, "TA B %c\r", m_bank);
	sendData(buf, 7);
	getNextLine(buf, 128);
	if (checkResponse(buf, "TA B ", "CUniden250::getBankTag()"))
	{
		if (buf[strlen(buf)-1]=='\r')
			buf[strlen(buf)-1]=0;
		m_banktag = buf+7;
	} else {
		return FALSE;
	}

	return TRUE;
}

void CUniden250::resetBankTag()
{
	// reset bank tag
	char buf[128];
	_snprintf(buf, 128, "TA B %c %s\r", m_bank, m_banktag);
	sendData(buf, (UINT)strlen(buf));
	getNextLine(buf, 128);
}

void CUniden250::setManualMode()
{
	UINT mode = 0;
	char buf[128];
	// check what mode we're in
	sendData("MD\r", 3);
	getNextLine(buf, 128);
	if (checkResponse(buf, "MD", "CUniden250::setFrequency()"))
	{
		buf[4]=0;
		mode = atoi(buf+2);
	}
	else
		return;

	// get to manual mode from wherever we are
	switch (mode)
	{
	// scan/manual mode, do nothing
	case 0:
	case 1:
		break;

	// trunking hold mode
	case 10:
	case 12:
	case 14:
	case 16:
	case 18:
	case 20:
		// press SEARCH key
		_snprintf(buf, 128, "KEY%0.2u\r", getSearchKeyCode());
		sendData(buf, 6);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(SRCH)"))
			return;
		// intentionally fall through to RESUME keypress

	// trunking scan/search mode
	case 9:
	case 11:
	case 13:
	case 15:
	case 17:
	case 19:
		// press RESUME key
		_snprintf(buf, 128, "KEY%0.2u\r", getResumeKeyCode());
		sendData(buf, 6);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(RSM)"))
			return;

		// intentionally fall through to MAN keypress

	// scan/search/VFO mode
	default:
		// press MAN key
		_snprintf(buf, 128, "KEY%0.2u\r", getManKeyCode());
		sendData(buf, 6);
		getNextLine(buf, 128);
		if (!checkResponse(buf, "OK", "CUniden250::setFrequency(MAN)"))
			return;
		break;
	}
}


BOOL CUnidenDynamic::initReceiver()
{
	/* get model info
	char buf[128];
	sendData("MDL\r", 4);
	getNextLine(buf, 128, "MDL");
	checkResponse(buf, "MDL", "CUnidenDynamic::initReceiver()");
	*/
	return TRUE;
}

void CUnidenDynamic::setFrequency(DOUBLE freq, BOOL narrow)
{
	// set VFO frequency
	char buf[128];
	_snprintf(buf, 128, "QSH,%08.0f,%04.0u,%s,0,0,0,0,0000000000000000,0,0\r", freq*10000, narrow?1250:2500, narrow?"NFM":"FM");
	sendData(buf, (UINT)strlen(buf));
	getNextLine(buf, 128, "QSH");
	checkResponse(buf, "QSH,OK", "CUnidenDynamic::setFrequency()");
}

DOUBLE CUnidenDynamic::getFrequency()
{
	DOUBLE freq = 0;
	char buf[128];

	// try GLG command
	sendData("GLG\r", 4);
	getNextLine(buf, 128, "GLG");
	if (buf[3]==',' && isdigit(buf[4]))
	{
		buf[12]=0;
		freq = ((DOUBLE)atoi(buf+4)/10000);

	} else if (buf[3]==',' && buf[4]!=',') {
		checkResponse(buf, "GLG response", "CUnidenDynamic::getFrequency()");
	}

	return freq;
}

BOOL CUnidenDynamic::isSquelchOpen()
{
	BOOL open = FALSE;
	char buf[128];

	// try GLG command
	sendData("GLG\r", 4);
	getNextLine(buf, 128, "GLG");
	if (buf[3]==',' && isdigit(buf[4]))
	{
		// scan string for squelch indicator
		char *cp = buf;
		for (int i = 0; i<8 && cp && (cp-buf)<128; ++i)
			cp = strchr(cp+1, ',');
		if (cp && cp[1]=='1')
			open = TRUE;

	} else if (buf[3]==',' && buf[4]!=',') {
		checkResponse(buf, "GLG response", "CUnidenDynamic::isSquelchOpen()");
	}

	return open;
}

void CUnidenDynamic::startScan()
{
	// press SCAN/SEARCH key
	char buf[64];
	sendData("KEY,S,P\r", 8);
	getNextLine(buf, 64, "KEY");
	checkResponse(buf, "KEY,OK", "CUnidenDynamic::startScan()");
}

void CUnidenDynamic::setBacklight(CReceiverCtrl::BacklightState state)
{
	// enter programming mode
	char buf[64];
	sendData("PRG\r", 4);
	getNextLine(buf, 64, "PRG");
	if (!checkResponse(buf, "PRG,OK", "CUnidenDynamic::setBacklight()"))
		return;

	// set backligt intensity
	_snprintf(buf, 64, "BKL,%s,RED\r", (state==High)?"HIGH":((state==Low)?"MIDL":"OFF"));
	sendData(buf, (UINT)strlen(buf));
	getNextLine(buf, 64, "BKL");
	if (!checkResponse(buf, "BKL,OK", "CUnidenDynamic::setBacklight()"))
		return;

	// exit programming mode
	sendData("EPG\r", 4);
	getNextLine(buf, 64, "EPG");
	if (!checkResponse(buf, "EPG,OK", "CUnidenDynamic::setBacklight()"))
		return;
}