/* $Id: nmt450.cpp,v 1.15 2007/01/11 16:57:04 olle Exp $
 *
 * Copyright (c) 2006, olle@toolcrypt.org
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 *    * Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *    * Redistributions in binary form must reproduce the above copyright notice,
 *      this list of conditions and the following disclaimer in the documentation
 *      and/or other materials provided with the distribution.
 *    * Neither the name of the <ORGANIZATION> nor the names of its contributors may
 *      be used to endorse or promote products derived from this software without 
 *      specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,DATA,
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "stdafx.h"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include "NMT450.h"

#include "cmdWndProt.h"
#include "callinfo.h"

// handle silly numerical values in subscriber numbers
#define HEX2NUM(x) (((x)==0)?10:(((x)==0xA)?0:x))

// re-mangle channel numbers to work as offsets from extended band
#define CHN2OFF(x) ((x>180)?x-180:x+20)

void
CNMT450::reportAddr(UINT prefix, UINT ident)
{
	// report new addresses
	snabel_addr_info_t *addr = new snabel_addr_info_t;					
	addr->system = SYSTEM_NMT450;
	addr->fleet = prefix;
	addr->addr = ident;
	CMDWND_REPORT_ADDR(cmdWnd, addr);
}

BOOL
CNMT450::getbit(int n, unsigned short *codewords)
{
    unsigned short mask;
    unsigned short codeword;

    codeword = codewords[n/16];
    mask = 1<<(15-(n%16));

    return (codeword & mask)?1:0;
}

#define POSSIBLE_TRANS(state, newstate) (((state<<1)&0x6)==(newstate&0x6))
#define OUTPUT(state, input)	(((((state>>1)&0x2)^(input<<1))^0x2) + (input&0x1))
#define INPUT(state)	(state&0x1)

int
CNMT450::simpleViterbi(unsigned short *sys, unsigned short *ecc, unsigned short *out)
{
	unsigned char pred[68][8];
	unsigned char metrics[68][8];
	unsigned char backtrace[68];

	// setup initial state
	for (int i=0; i<68; i++)
		for (int ii=0; ii<8; ii++)
			metrics[i][ii] = 0xff;
    metrics[0][0] = 0;

	// process 64 symbols of 2 bits each
	for (int i=1; i<68; i++)
	{
		// get the symbol
		unsigned char sym = 0;
		sym += (getbit(i, ecc)&0x1);
		sym <<= 1;
		sym += (i<64)?(getbit(i, sys)&0x1):0;

		// calculate state transitions
		for (unsigned char laststate=0; laststate<8; laststate++)
			if (metrics[i-1][laststate]!=0xff)
			{
				for (unsigned char state=0; state<8; state++)
					if (POSSIBLE_TRANS(laststate, state))
					{
						// calculate accumulated hamming distance for this path
						unsigned char metric = metrics[i-1][laststate] + ((OUTPUT(laststate, INPUT(state))&0x3) ^ sym);
						if (metrics[i][state] > metric)
						{
							metrics[i][state] = metric;
						
							// save predecessor info
							pred[i][state] = laststate;
						}
					}
			}
	}
	
#ifdef DEBUG_FEC
	for (unsigned char row=0; row<8; row++)
	{
		CString dbg;
		dbg.Format("%.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x %.2x", metrics[45][row], metrics[46][row], metrics[47][row], metrics[48][row], metrics[49][row], metrics[50][row], metrics[51][row], metrics[52][row], metrics[53][row], metrics[54][row], metrics[55][row], metrics[56][row], metrics[57][row], metrics[58][row], metrics[59][row], metrics[60][row], metrics[61][row], metrics[62][row], metrics[63][row], metrics[64][row]);
		CMDWND_ADD_LOGSTR(mmcmdWnd, dbg);
	}
#endif

	int nerrors = -1;

	// get final state with least accumulated errors
	for (unsigned char state=0; state<8; state++)
		if (metrics[67][state] < (unsigned int)nerrors)
		{
			nerrors = metrics[67][state];
			backtrace[67] = state;
		}

#ifdef DEBUG_FEC
	CString dbg;
	dbg.Format("final state = %x metric = %d", backtrace[64], nerrors);
	CMDWND_ADD_LOGSTR(mmcmdWnd, dbg);
#endif

	// do simple viterbi backtrace
	for (int i=67; i>0; i--)
		backtrace[i-1] = pred[i][backtrace[i]];
	
	// decode the state symbols into output buffer
	for (int i=0; i<64; i++)
	{
		out[0] <<= 1;
		out[0] += (out[1] & 0x8000)?1:0;
		out[1] <<= 1;
		out[1] += (out[2] & 0x8000)?1:0;
		out[2] <<= 1;
		out[2] += (out[3] & 0x8000)?1:0;
		out[3] <<= 1;
		if (INPUT(backtrace[i]))
			out[3] += 1;
	}
	
	return nerrors;
}


void
CNMT450::handleBit(bool bit)
{
    // Shift in the new bit into the 160 bit buffer.
    codeword[0] <<= 1;
    codeword[0] += (codeword[1] & 0x80000000)?1:0;
    codeword[1] <<= 1;
    codeword[1] += (codeword[2] & 0x80000000)?1:0;
    codeword[2] <<= 1;
    codeword[2] += (codeword[3] & 0x80000000)?1:0;
    codeword[3] <<= 1;
    codeword[3] += (codeword[4] & 0x80000000)?1:0;
    codeword[4] <<= 1;
    if (bit)
		codeword[4] += 1;

	// Update bitcount
	if (cnt)
		cnt--;

	// check for MUTE sequence '101010101'+'11100010010'+'x0x0x0x0x0x0'
	//                          bit-sync    frame-sync    frame-ident
	if ((codeword[4] & 0xFFFFF555) == 0xAAF12000)
	{
		// if we don't have sync, report found sync
		if(!synctimer)
			CMDWND_GOT_SYNC(cmdWnd);

		// reset sync timer
		synctimer = 1;

		// set number of bits til next frame
		cnt = 129;
	} else {
		// if synched, tick the synctimer
		if(synctimer && synctimer++ > 250)
		{
			// report lost sync after 250 bits
			CMDWND_LOST_SYNC(cmdWnd);
			synctimer = 0;
		}
	}


	// check for full frame arrival '101010101'+'11100010010'
	//                               bit-sync    frame-sync
	if (((codeword[0] & 0xFFFFF000) == 0xAAF12000) || cnt==1)
	{
		unsigned short frame[4];
		unsigned short data[4];
		unsigned short err[5];
		memset(frame, 0, sizeof(frame));
		memset(data, 0, sizeof(data));
		memset(err, 0, sizeof(err));

		// de-interlace the 64 data bits
		for (int nword=1; nword<5; nword++)
			for (int i=15; i>=0; i--)
			{
				data[0] <<= 1;
				data[0] += (data[1] & 0x8000)?1:0;
				data[1] <<= 1;
				data[1] += (data[2] & 0x8000)?1:0;
				data[2] <<= 1;
				data[2] += (data[3] & 0x8000)?1:0;
				data[3] <<= 1;
				if ((codeword[nword] >> i*2) & 0x1)
					data[3] += 1;
			}

		// de-interlace the 67 error correction bits
		int nbit = 0;
		for (int nword=0; nword<6; nword++)
			for (int i=16; i>0; i--)
			{
				// only get last 3 bits of first codeword
				if (nword==0 && i>6)
					i=6;

				// fill 80 bits buffer
				if (nbit++>79)
					break;

				// shift it into result
				err[0] <<= 1;
				err[0] += (err[1] & 0x8000)?1:0;
				err[1] <<= 1;
				err[1] += (err[2] & 0x8000)?1:0;
				err[2] <<= 1;
				err[2] += (err[3] & 0x8000)?1:0;
				err[3] <<= 1;
				err[3] += (err[4] & 0x8000)?1:0;
				err[4] <<= 1;
				if ((codeword[nword] >> ((i*2)-1)) & 0x1)
					err[4] += 1;
			}

		// decode convoluted ECC using simple viterbi algorithm
		int errors = simpleViterbi(data, err, frame);

		// accept some corrected bits, otherwise frame is unrecoverable
		if (errors>6)
		{
#ifdef DEBUG
			// calculate S/N ratio for frames
			double snr = (double)goodframes/(double)badframes;

			CString dbg;
			dbg.Format("ERROR| Bad frame. %2u errors detected.  | %04x %04x %04x %04x  S/N: %2.3f", errors, data[0], data[1], data[2], data[3], snr);
			CMDWND_ADD_LOGSTR(cmdWnd, dbg);
			
			if (badframes++ == 0)
				badframes = -1000;
#endif			
			return;
		}
#ifdef DEBUG
		if (goodframes++ == 0)
			goodframes = -1000;
#endif			

		// set number of bits til next frame
		cnt = 140+15+11+1;

		// parse frame contents
		unsigned short channel = ((frame[0]>>4)&0xf) + (((frame[0]>>8)&0xf)*10) + (((frame[0]>>12)&0x1)*100);
		if (!channel)
			channel = 200;
		unsigned char prefix = (frame[0]&0xf);
		unsigned char TA = HEX2NUM((frame[1]>>12)&0xf)*10 + HEX2NUM((frame[1]>>8)&0xf);

		// subscriber number
		UCHAR Z = (frame[1]>>4)&0xf;
		UINT X = HEX2NUM(frame[1]&0xf)*100000 + HEX2NUM((frame[2]>>12)&0xf)*10000 + HEX2NUM((frame[2]>>8)&0xf)*1000 +
				 HEX2NUM((frame[2]>>4)&0xf)*100 + HEX2NUM(frame[2]&0xf)*10 + HEX2NUM((frame[3]>>12)&0xf);

#ifdef DEBUG
		if (prefix != 4)
		{
			CString str;
			str.Format("DEBUG| Ch: %3u Pfx: %2u TA: %2u EC:%2u    | %04x %04x %04x %04x (%04x %04x %04x %04x)", channel, prefix, TA, errors, frame[0], frame[1], frame[2], frame[3], data[0], data[1], data[2], data[3]);
			CMDWND_ADD_LOGSTR(cmdWnd, str);
		}
#endif

		// parse frame type
		switch (prefix)
		{	
		case  0: // Frame 6  - Idle Frame
#ifdef DEBUG
			CMDWND_ADD_LOGSTR(cmdWnd, "IDLE | Idle Frame                      |");
#endif
			break;

		case  3: // Frame 4  - Free traffic channel indication
			{
			CString tcind;
			tcind.Format("TCIND| Free traffic channel indication | Ch(%3u) TA(%2u)", channel, TA);
			CMDWND_ADD_LOGSTR(cmdWnd, tcind);
			}
			break;

		case  4: // Frame 1b - Combined calling and traffic channel indication

			// if entering new TA or switched calling channel, report
			if (channel!=currentchannel || TA!=currentTA)
			{
				// send system report
				reportAddr(0, 0);

				// save channel/TA
				currentchannel = channel;
				currentTA = TA;

#ifdef DEBUG
				// clear counters
				goodframes = badframes = 1;
#endif

				// report new channel/TA
				CString report;
				report.Format("CTIND| Calling/Traffic Ch. indication  | Ch(%3u) TA(%2u)", channel, TA);
				CMDWND_ADD_LOGSTR(cmdWnd, report);
			}
			break;

		case  5: // Frame 3a - Traffic channel allocation on traffic channel
			if (Z!=15) // Z(15) == MTX->BS
			{
				USHORT N = (frame[3]&0xf) + (((frame[3]>>4)&0xf)*10) + (((frame[3]>>8)&0x1)*100);
				if (!N)
					N = 200;

				if ((frame[3]&0xfff)==0x000) // Frame 3b - Identity request on traffic channel
				{
					CString allct;
					allct.Format("IDREQ| Identity request on traffic ch. | Ch(%3u) TA(%2u)  Z(%2u) X(%06u)", channel, TA, Z, X);
					CMDWND_ADD_LOGSTR(cmdWnd, allct);
				} else {
					// report new call
					snabel_call_info_t* call = new snabel_call_info_t;
					call->ctrlchnl = CHN2OFF(currentchannel);
					call->calladdr.system = call->fromaddr.system = SYSTEM_NMT450;
					call->calladdr.fleet = call->fromaddr.fleet = Z;
					call->calladdr.addr = X;
					call->fromaddr.addr = 0;
					call->callchnl = CHN2OFF(N);
					CMDWND_REPORT_CALL(cmdWnd, call);
					
					CString allct;
					allct.Format("ALLCT| Traffic channel allocation      | Ch(%3u) TA(%2u)  Z(%2u) X(%06u) N(%3u)", channel, TA, Z, X, N);
					CMDWND_ADD_LOGSTR(cmdWnd, allct);
				}
			}  else {
				CString sstr;
				sstr.Format("SIGNS| Signal strength measurmnt order | Ch(%3u) TA(%2u)  Z(%2u)", channel, TA);
				CMDWND_ADD_LOGSTR(cmdWnd, sstr);
			}
			break;

		case  6: // Frame 5a - Line signal
			switch(frame[3]&0xfff) // L(???)
			{
			case 0x333: // L( 3) - Proceed to send unencrypted digits.
			case 0xbbb: // L(11) - Proceed to send encrypted digits.
				{
				CString str;
				str.Format("LINES| Line signal - Ready to dial     | Ch(%3u) TA(%2u)  Z(%2u) X(%06u)", channel, TA, Z, X);
				CMDWND_ADD_LOGSTR(cmdWnd, str);
				}
				break;
			case 0xddd: // L(13) - Clearing, call transfer activated.
			case 0xfff: // L(15) - Clearing, call transfer not activated.
				if ((frame[2]&0xffff)!=0xffff)
				{
					// report call cleardown
					snabel_call_info_t* call = new snabel_call_info_t;
					call->ctrlchnl = CHN2OFF(currentchannel);
					call->calladdr.system = call->fromaddr.system = SYSTEM_NMT450;
					call->calladdr.fleet = Z;
					call->calladdr.addr = X;
					call->callchnl = CHN2OFF(channel);
					CMDWND_CLEAR_CALL(cmdWnd, call);
				
					CString clear;
					clear.Format("LINES| Line signal - Clearing channel  | Ch(%3u) TA(%2u)  Z(%2u) X(%06u) L(%2u)", channel, TA, Z, X, frame[3]&0xf);
					CMDWND_ADD_LOGSTR(cmdWnd, clear);
				} else {
					CString unkn;
					unkn.Format("UNKWN| Ch: %3u Prefix: %2u TA: %2u EC:%2u | %04x %04x %04x %04x", channel, prefix, TA, errors, frame[0], frame[1], frame[2], frame[3]);
					CMDWND_ADD_LOGSTR(cmdWnd, unkn);
				}
				break;

			case 0x666: // L(6) - Address complete
				{
				// report new call
				snabel_call_info_t* call = new snabel_call_info_t;
				call->ctrlchnl = CHN2OFF(currentchannel);
				call->calladdr.system = call->fromaddr.system = SYSTEM_NMT450;
				call->calladdr.fleet = call->fromaddr.fleet = Z;
				call->calladdr.addr = 0;
				call->fromaddr.addr = X;
				call->callchnl = CHN2OFF(channel);
				CMDWND_REPORT_CALL(cmdWnd, call);
				
				CString allct;
				allct.Format("LINES| Line signal - Address Complete  | Ch(%3u) TA(%2u)  Z(%2u) X(%06u)", channel, TA, Z, X);
				CMDWND_ADD_LOGSTR(cmdWnd, allct);
				}
				break;

			case 0x999: // L(9) - Ringing order
				{
				CString allct;
				allct.Format("LINES| Line signal - Ringing order     | Ch(%3u) TA(%2u)  Z(%2u) X(%06u)", channel, TA, Z, X);
				CMDWND_ADD_LOGSTR(cmdWnd, allct);
				}
				break;

			default:
				CString clear;
				clear.Format("LINES| Line signal                     | Ch(%3u) TA(%2u)  Z(%2u) X(%06u) L(%2u)", channel, TA, Z, X, frame[3]&0xf);
				CMDWND_ADD_LOGSTR(cmdWnd, clear);
				break;
			}
			break;

		case  8: // Frame 7  - Authentication request
			{
			CString str;
			str.Format("AUTHR| Authentication request          | Ch(%3u) TA(%2u)  H(%.2x%.4x%.1x)", channel, TA, frame[1]&0xff, frame[2], (frame[3]>>12)&0xf);
			CMDWND_ADD_LOGSTR(cmdWnd, str);
			}
			break;

		case  9: // Frame 3c - Traffic channel allocation on traffic channel, short procedure
			{
				USHORT N = (frame[3]&0xf) + (((frame[3]>>4)&0xf)*10) + (((frame[3]>>8)&0x1)*100);
				if (!N)
					N = 200;

				if (N==channel)
				{
					// report new call
					snabel_call_info_t* call = new snabel_call_info_t;
					call->ctrlchnl = CHN2OFF(currentchannel);
					call->calladdr.system = call->fromaddr.system = SYSTEM_NMT450;
					call->calladdr.fleet = call->fromaddr.fleet = Z;
					call->calladdr.addr = X;
					call->fromaddr.addr = 0;
					call->callchnl = CHN2OFF(N);
					CMDWND_REPORT_CALL(cmdWnd, call);
					
					CString allct;
					allct.Format("ALLCT| Traffic channel allocation      | Ch(%3u) TA(%2u)  Z(%2u) X(%06u) N(%3u)", channel, TA, Z, X, N);
					CMDWND_ADD_LOGSTR(cmdWnd, allct);
				} else {
					CString unkn;
					unkn.Format("UNKWN| Ch: %3u Prefix: %2u TA: %2u EC:%2u | %04x %04x %04x %04x", channel, prefix, TA, errors, frame[0], frame[1], frame[2], frame[3]);
					CMDWND_ADD_LOGSTR(cmdWnd, unkn);
				}
			}
			break;

					// Frame 1a - Calling channel indication
					// Frame 2a - Call to mobile subscriber on calling channel	
		case 12:	// Frame 2b - Traffic channel allocation on calling channel
			switch(frame[3]&0xfff)
			{
			case 0x000: // N(0) - Frame 1a/2a
				if (Z && Z<14)
				{
					CString dbg;
					dbg.Format("CALLM| Call to mobile subscriber       | Ch(%3u) TA(%2u)  Z(%2u) X(%06u)", channel, TA, Z, X);
					CMDWND_ADD_LOGSTR(cmdWnd, dbg);

					// report address
					reportAddr(Z, X);
					break;

				} else if (Z==14) { // Z(14) - Additional info (battery saving)
					CString str;
					str.Format("BATTS| Battery saving order            | Ch(%3u) TA(%2u)", channel, TA);
					CMDWND_ADD_LOGSTR(cmdWnd, str);
					break;

				} else if (Z==15) { // Z(15) - Information to/from BS
					CString str;
					str.Format("UNKWN| Information to/from BS          | Ch(%3u) TA(%2u)", channel, TA);
					CMDWND_ADD_LOGSTR(cmdWnd, str);
					break;

				} else if ((frame[2]&0xffff)==0x0000) {
					// if entering new TA or switched calling channel, report
					if (channel!=currentchannel || TA!=currentTA)
					{
						// send system report
						reportAddr(0, 0);

						// save channel/TA
						currentchannel = channel;
						currentTA = TA;

#ifdef DEBUG
						// clear counters
						goodframes = badframes = 1;
#endif

						// report new channel/TA
						CString report;
						report.Format("CCIND| Calling channel indication      | Ch(%3u) TA(%2u)", channel, TA);
						CMDWND_ADD_LOGSTR(cmdWnd, report);
					}
					break;
				}

			default:
				CString unkn;
				unkn.Format("UNKWN| Ch: %3u Prefix: %2u TA: %2u EC:%2u | %04x %04x %04x %04x", channel, prefix, TA, errors, frame[0], frame[1], frame[2], frame[3]);
				CMDWND_ADD_LOGSTR(cmdWnd, unkn);
				break;
			}
			break;

		case 15: // Frame 20 - Channel activation order
			if (Z==15) // Z(15) == MTX->BS order
			{
				BOOL clear = FALSE;
				switch((frame[2]>>4)&0xf) // A(?)
				{
				case 0x0: // A(0)  - Idle radio channel
					CMDWND_ADD_LOGSTR(cmdWnd, "CORDR| Channel activation order        | Idle radio channel");
					clear = TRUE;
					break;
				case 0x2: // A( 2) - General channel reset
					CMDWND_ADD_LOGSTR(cmdWnd, "CORDR| Channel activation order        | General channel reset");
					clear = TRUE;
					break;
				case 0xc: // A(12) - Stop sending -signal, etc.
					CMDWND_ADD_LOGSTR(cmdWnd, "CORDR| Channel activation order        | Stop sending -signal");
					clear = TRUE;
					break;
				case 0xd: // A(13) - Start sending -signal, etc.
					CMDWND_ADD_LOGSTR(cmdWnd, "CORDR| Channel activation order        | Start sending -signal");
					break;
				default:
					CString cordr;
					cordr.Format("CORDR| Channel activation order        | A: %2u", (frame[2]>>4)&0xf);
					CMDWND_ADD_LOGSTR(cmdWnd, cordr);
					break;
				}
				if (clear)
				{
					// report call cleardown
					snabel_call_info_t* call = new snabel_call_info_t;
					call->ctrlchnl = CHN2OFF(currentchannel);
					call->calladdr.system = call->fromaddr.system = SYSTEM_NMT450;
					call->callchnl = CHN2OFF(channel);
					CMDWND_CLEAR_CALL(cmdWnd, call);
				}
				break;
			}
			//break;

		default:
			CString unkn;
			unkn.Format("UNKWN| Ch: %3u Prefix: %2u TA: %2u EC:%2u | %04x %04x %04x %04x", channel, prefix, TA, errors, frame[0], frame[1], frame[2], frame[3]);
			CMDWND_ADD_LOGSTR(cmdWnd, unkn);
			break;
		}
	}
}
